# Copyright 2018 - 2019 Ori Ben-Moshe - All rights reserved.
from threading import Thread
import cv2


class Camera:
    """
    A camera object used to take images in a higher fps than cv2.VideoCapture
    """
    def __init__(self, src=0, width=320, height=240):
        self.stream = cv2.VideoCapture(src)
        self.stream.set(3, width)
        self.stream.set(4, height)
        (self.grabbed, self.frame) = self.stream.read()
        self.stopped = False

    def start(self):
        # start the thread to read frames from the video stream
        Thread(target=self.update, args=()).start()
        return self

    def update(self):
        while True:
            if self.stopped:
                return
            (self.grabbed, self.frame) = self.stream.read()

    def read(self):
        """
        Action: Returns the ret val and the frame
        :return: return val (if image taking was successful), the image numpy
        :rtype: bool, numpy.array
        """
        return self.grabbed, self.frame

    def stop(self):
        """
        Stops the camera thread.
        :return:
        """
        self.stopped = True

    def release(self):
        """
        Action: stops the camera thread and releases the camera
        :return: None
        """
        self.stopped = True
        self.stream.release()

    def set(self, prop_id, value):
        """
        Action: sets the value of the given property to the given value
         :param prop_id: the proprety number, for more information:
                        https://docs.opencv.org/3.1.0/d8/dfe/classcv_1_1VideoCapture.html
        :param value: the value to be set, a number
        :return: None
        """
        self.stream.set(prop_id, value)

    def get(self, prop_id):
        """
        Action: retrieves the value of a property based on its id
        :param prop_id: the proprety number, for more information:
                        https://docs.opencv.org/3.1.0/d8/dfe/classcv_1_1VideoCapture.html
        :return: the value
        """
        return self.stream.get(prop_id)

    def isOpened(self):
        """
        Returns true if the camera is opened and false otherwise
        :return:
        :rtype: bool
        """
        return self.isOpened()

    def getBackendName(self):
        """
        Action: returns the Back end name of a camera
        :return:
        """
        return self.stream.getBackendName()



from . import Vision
from . import Color
from . import General
from . import Calibration
from . import Drawing
from . import Geometry
from . import Sorters
from . import Filters
from . import Directions
from . import ImageFilters

__name__ = "Ovl"
__version__ = '0.2.4.1'



# Copyright 2018 - 2019 Ori Ben-Moshe - All rights reserved.
from socket import *
from numpy import copy, vstack, hstack, ndarray
from .Color import Color, BuiltInColors, MultiColor
from .General import *
from . import Sorters
import time
import copy
from inspect import getmembers, isfunction
import cv2
import json
from networktables import NetworkTables
from sys import version_info
from .Camera import Camera
try:
    from serial import Serial
    serial_exists = True
except ImportError:
    print('The serial module (pyserial) is not installed (or found),\n Vision Serial connection is disabled.')
    serial_exists = False


if version_info[0] == 3:
    xrange = range


def display_image(image, win_name='image', resizable=False):
    """
    Action: The function displays the image whose path is received.
    :param image: Represents an image path (string), an already open image in the for of a numpy array (ndarray)
                  or a list of images
    :param win_name: Name of the Window that displays the images
    :param resizable: A boolean that determines whether the image window is resizable or not
    :return: None
    """
    if type(image) is str:
        img = cv2.imread(image)
        if resizable:
            cv2.namedWindow(win_name, cv2.WINDOW_NORMAL)
        cv2.imshow(win_name, img)
        cv2.waitKey()
    elif type(image) is ndarray:
        if resizable:
            cv2.namedWindow(win_name, cv2.WINDOW_NORMAL)
        cv2.imshow(win_name, image)
        cv2.waitKey()
    if type(image) is list:
        if type(image[0]) is str:
            image[0] = cv2.imread(image[0])
        shape = image[0].shape
        width = shape[0]
        height = shape[1]
        blob = image[0]
        for img in image[1:]:
            if type(img) is str:
                img = cv2.imread(img)
            im_shape = img.shape
            if width + im_shape[0] > height + im_shape[1]:
                blob = vstack((blob, img))
                width = im_shape[0]
            else:
                blob = hstack((blob, img))
                height += im_shape[1]
        if resizable:
            cv2.namedWindow(win_name, cv2.WINDOW_NORMAL)
        cv2.imshow(win_name, blob)
        cv2.waitKey()


class InvalidImage(object):

    def __init__(self):
        self.false = False

    def __bool__(self):
        return self.false


def get_all_functions():
    """
    Action: Returns a list of all available functions within the current scope
    :return: list of function object that are callable
    """
    f_list = []
    for module, module_object in globals().items():
        for sub_object_name, sub_object in getmembers(module_object):
            if isfunction(sub_object):
                f_list.append((sub_object_name, sub_object))
                continue
            for sub_members_name, sub_members_objects in getmembers(sub_object):
                if isfunction(sub_members_objects):
                    f_list.append((sub_members_name, sub_members_objects))
    return f_list


class Vision(object):
    """
    The vision object represents a control object to the whole process of the vision processing
    it receives  a list of filter functions, a directions function, the connection destination address,
    a port (or key) to connect to, list of parameters for the filter functions and 2 LinearDiscriminantAnalysis vectors.
    """
    NO_CONNECTION = 'NOCONNECTION'
    NT_CONNECTION = 'NT'  # Network Tables via pynetworktables
    BT_CONNECTION = 'BT'  # BlueTooth via Socket, Not Implemented
    UDP_CONNECTION = 'UDP'  # UDP via Socket  
    TCP_CONNECTION = 'TCP'  # TCP via Socket, Not Implemented
    SERIAL_CONNECTION = 'SP'  # Serial Port via

    list_of_functions = get_all_functions()

    @staticmethod
    def is_static_ip(ip):
        """
        Action: checks if the given ip is a static ip for FRC
        :param ip: the ip to be checked, a str
        :return: True if the ip is an FRC static ip, False if it isn't
        """
        if type(ip) != str:
            return False
        if len(ip) == 10:
            if ip.startswith('10.') and ip.endswith('.2') and ip.count('.') == 3:
                return True
        return False

    @property
    def contour_amount(self):
        return len(self.contours)

    class FilterReplacement:
        def __init__(self, name):
            self.name = name

        def __name__(self):
            return self.name

    def get_filters(self):
        """
        Action: returns the list of filter functions
        :return:
        """
        return self.__filters

    def get_params(self):
        """
        Action: returns the list of the parameters for the filter functions
        :return:
        """
        return self.__params

    @staticmethod
    def is_valid_ip(ip):
        """
        Action: checks if the given ip address (string) is a valid ip
        :param ip: the string of the ip to be checked, can be ipv6 and ipv4
        :return: False if invalid, True if valid
        """
        try:
            inet_aton(ip)
            return True
        except error:
            return False

    def __init__(self,
                 color=None,
                 filters=None,
                 directions_function=None,
                 target_amount=1,
                 width=320,
                 height=240,
                 connection_dst=NO_CONNECTION,
                 port=None,
                 log_file=None,
                 camera_port=None,
                 failed_value='9999',
                 **kwargs):
        """
        Action: The vision object represents a control object to the whole process of the vision processing
        it receives  a list of filter functions, a directions function, the connection destination address,
        a UDP port to connect to, list of parameters for the filter functions and 2 vectors.
        :param filters: the list of filter functions
        :param directions_function: a functions that receives a list or a single contour and returns directions
        :param connection_dst: the connection (ip or hostname) in the network to connect to.
        :param port: the port number for the connection.
        :param log_file: a path to a file where output should be displayed
        :param connection_type: The type of connection you are using to pass the data Vision.NT
        :param kwargs: The optional key word arguments are: 
                       parameters for specific parameters for filter functions,
        """
        if log_file is not None:
            self.log_path = log_file
            if type(self.log_path) not in (str, bool, type(None)):
                raise TypeError('Invalid log file type, must be a file path, bool or None')
            if self.log_path is False:
                self.log_path = None
        else:
            self.log_path = None

        self.width = width
        self.height = height
        if type(width) != int:
            raise TypeError('The width of images taken must be an integer')
        if type(height) != int:
            raise TypeError('The width of images taken must be an integer')
            
        if filters is None:
            self.__filters = []
        else:
            self.__filters = filters
            if len(filters) == 0:
                print_pipe(self.log_path, "No filters given, Using raw contour result.")
        if "parameters" in kwargs:
            self.__params = kwargs['parameters']
        elif "params" in kwargs:
            self.__params = kwargs['params']
        else:
            self.__params = None

        target_amount = int(target_amount)
        if type(target_amount) is not int:
            self.target_amount = 1
        elif target_amount < abs(target_amount):
            self.target_amount = 0
        else:
            self.target_amount = target_amount

        self.contours = []

        self.camera = None
        self.camera_port = None
        camera_port_type = type(camera_port)
        if camera_port_type is cv2.VideoCapture:
            self.camera = camera_port
        elif camera_port_type is Camera:
            self.camera = camera_port
        elif camera_port is None:
            pass
        else:
            use_camera = kwargs['use_camera'] if 'use_camera' in kwargs else False
            self.camera_setup(camera_port, width, height, camera=use_camera)

        # save the direction function
        self.directions = directions_function
        
        if type(color) is Color:
            self.color = color
            self.low = color.low
            self.high = color.high
        elif type(color) is MultiColor:
            self.color = color
        elif type(color) in (tuple, list, set):
            self.color = color
        else:
            self.color = False
            
        # Calibration configuration
        self.json_file = None
        if 'calibration' in kwargs:
            cal_file = kwargs['calibration']
            self.json_file = cal_file
            if type(cal_file) is str:
                if not cal_file.endswith('.json'):
                    raise ValueError('Calibration file is not a json file')
                with open(cal_file, 'r') as calibration_file:
                    cal = json.loads(calibration_file.read())
                    self.saturation_weight_vector = cal['swv']
                    self.value_weight_vector = cal['vwv']
        else:
            self.value_weight_vector = None
            self.saturation_weight_vector = None
            if 'svw' in kwargs:
                self.saturation_weight_vector = kwargs['swv']

            if 'vwv' in kwargs:
                    self.value_weight_vector = kwargs['vwv']
                    
        # Value returned if failed to find the target.
        self.failed_value = failed_value
        if type(self.failed_value) is int:
            self.failed_value = str(self.failed_value)

        # Network related configurations
        self.raw_address = connection_dst
        self.network_port = port
        if connection_dst == Vision.NO_CONNECTION:
            self.connection_address = None
            self.network_port = port
            self.socket = None
            self.connection_type = None
        elif Vision.is_static_ip(connection_dst):
                self.connection_address = connection_dst
                NetworkTables.initialize(server=connection_dst)
                self.socket = NetworkTables.getTable('SmartDashboard')
                self.connection_type = Vision.NT_CONNECTION
                self.network_port = port
        elif connection_dst == "@Serial" and serial_exists:
            if type(port) in (tuple, set, list):
                self.socket = Serial(port[0], baudrate=[1])
            elif type(port) is str:
                self.socket = Serial(port)
            else:
                raise TypeError
            self.connection_type = Vision.SERIAL_CONNECTION
            self.connection_address = port
            self.network_port = port
        elif Vision.is_valid_ip(connection_dst):
            self.network_port = port if 0 < port < 65535 else 0
            try:
                self.socket = socket()
                self.connection_address = connection_dst
                self.socket.bind((connection_dst, port))
            except Exception as e:
                print_pipe(self.log_path, 'Failed to connect.', e)
        else:
            self.network_port = port if 0 < port < 65535 else 0
            try:
                self.connection_address = gethostbyname(connection_dst)
                self.socket = socket()
                self.socket.bind((connection_dst, port))
            except gaierror:
                self.connection_address = False
                raise ValueError('All connections failed, check the given address and if the connection is valid.')

        # Morphological functions and image filters
        self.image_filters = kwargs['image_filters'] if 'image_filters' in kwargs else []
        self.image_params = kwargs['image_params'] if 'image_params' in kwargs else [] * len(self.image_filters)

        self.morph = kwargs['morph'] if 'morph' in kwargs else []
        self.morph_params = kwargs['morph_params'] if 'morph_params' in kwargs else [] * len(self.morph)

    def apply_morphs(self, mask):
        """
        Action: Applies all morphological functions on the mask
        :param mask: the mask on which the functions should be applied
        :return: the applied mask
        """
        if self.morph not in (tuple, list, set):
            return mask
        for idx, mask_filter in enumerate(self.morph):
            mask = mask_filter(mask, *self.morph_params[idx])
        return mask

    def apply_sorter(self, sorter_function=Sorters.dec_area_sort):
        """
        Action: applies a sorter function to the contour list
        :param sorter_function: the sorter function to be applied
        :return: the contour list (list of numpy arrays)
        """
        self.contours = sorter_function(self.contours)
        return self.contours

    def send_to_destination(self, data):
        """
        Action: Sends data to the destination.
        :param data: The data to send to the RoboRIO - Usually rhe directions calculated by the direction function
        :return: True if data was successfully sent, False if not.
        """
        if self.connection_type == Vision.NT_CONNECTION:
            self.socket.putNumber(self.network_port, data)
        elif self.connection_type == Vision.UDP_CONNECTION:
            self.socket.sendto(str(data), (self.connection_address, self.network_port))
        elif self.connection_type == Vision.SERIAL_CONNECTION:
            self.socket.write(str(data))

    def get_image(self):
        """
        Action: gets an image from self.camera
        :return: the image, false if failed
        """
        if self.camera is None:
            raise ValueError("No camera connected")
        output = self.camera.read()
        if len(output) == 2:
            ret, img = output
            if ret:
                return img
            else:
                return False
        else:
            return output

    def get_applied_image(self):
        """
        Action: gets an image from self.camera and applies all image filters
        :return: the image filter applied image
        """
        output = self.get_image()
        return self.apply_image_filters(output) if output is not False else False

    def apply_filter(self, filter_function, apply_results=False, parameters=None, suppress=False):
        """
        Action: applies a filter function of the contour list
        :param filter_function: Filter functions are function with a contour list variable that apply some
        sort of filter on the contours, thus removing ones that don't fit the limit given by the filter.
        for example: straight_rectangle_filter removes contours that are not rectangles that are parallel to the edges
        of the picture.
        :param apply_results: If you want to apply a filter only to debug and only see its affects without it affecting
        the contour
        :param parameters: Additional parameters for the function, uses the repr of the each parameter, the additional
         is in addition to the.
        :param suppress: if true does not print anything
        :return: returns the output of the filter function.
        """
        if not suppress:
            string = 'Before "%": '.\
                         replace('%', filter_function.__name__.replace('_', ' ')) + str(self.contour_amount)
            print_pipe(self.log_path,
                       string)
        if parameters not in ((), [], None):
            o = filter_function(self.contours, *parameters)
        else:
            o = filter_function(self.contours)

        if type(o) is tuple:
            if len(o) == 2:
                output, ratio = o[0], o[1]
            else:
                raise InvalidCustomError('Filter function must return between 1 and 2 lists.'
                                         'Please refer to the Documentation: https://github.com/1937Elysium/Ovl-Python')
        elif type(o) in (list, ndarray):
            output, ratio = o, []
        else:
            raise TypeError('The contour list must be a list or a ndarray')

        if apply_results:
            if type(output) not in (list, ndarray):
                raise TypeError("Filter function must return a contour or a contour list!")
            self.contours = output
        if not suppress:
            print_pipe(self.log_path, 'After %: '.replace('%', filter_function.__name__) + str(len(output)))
        return output, ratio

    def apply_all_filters(self, apply_all=True, suppress=False):
        """
        Action: Applies all of the filters on the
        :param apply_all: A Boolean that determines whether or not apply all filters.
        :param suppress: Does not print before and after contour count if True.
        :return: a list of all of the ratios given by the filter function in order.
        """
        all_ratios = []
        for idx, filter_func in enumerate(self.__filters):
            if self.__params is not None and self.__params[idx] not in ((), []):
                _, ratio = self.apply_filter(filter_func,
                                             apply_results=apply_all,
                                             parameters=self.__params[idx],
                                             suppress=suppress)
                all_ratios.append(ratio)
            else:
                _, ratio = self.apply_filter(filter_func,
                                             apply_results=apply_all,
                                             suppress=suppress)
                all_ratios.append(ratio)
        return all_ratios

    def get_color_mask(self, img, color=None):
        """
        Action: gets a mask for a given img and color object(if no color object is given uses self.color
        :param img: the numpy array of the image
        :param color: the Color or MultiColor object used to create the binary mask
        :return: the binary mask
        """
        c = color if color is not None else self.color
        if type(c) not in (Color, MultiColor):
            raise TypeError("Color must be Color or MultiColor object!")
        img_in_hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
        if type(c) == MultiColor:
            img_mask = cv2.inRange(img_in_hsv, c.colors[0].low, c.colors[0].high)
            for color_range in c.colors[1:]:
                img_mask += cv2.inRange(img_in_hsv, color_range.low, color_range.high)
        else:
            img_mask = cv2.inRange(img_in_hsv, c.low, c.high)
        return img_mask

    def get_contours_mask(self, mask, save=True, return_hierarchy=False, apply_morphs=True):
        """
        Actions: gets contours from the given mask and apply
        :param mask: binary image (mask), a numpy array
        :param save: if the contours should be saved to self.contours
        :param return_hierarchy: if the hierarchy should be returned
        :param apply_morphs: if the morphological functions should be applied.
        :return: the list of contours
        """
        mask = self.apply_morphs(mask) if apply_morphs else mask
        result = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        if len(result) == 3:
            found_contours = result[1]
            hierarchy = result[2]
            if save:
                self.contours = found_contours
            return found_contours, hierarchy if return_hierarchy else found_contours
        elif len(result) == 2:
            found_contours = result[0]
            hierarchy = result[1]
            if save:
                self.contours = found_contours
            return found_contours, hierarchy if return_hierarchy else found_contours
        else:
            raise ValueError("Didn't get 2 or 3 out puts from findContours")

    def get_edge_contours(self, img, values=None, save=True, return_hierarchy=False):
        """
        Actions: gets contours from the given img using the Canny function
        :param img: image to be searched, a numpy array
        :param values: parameters other than the image to be passed to the function.
        :param save: if the contours should be saved to self.contours
        :param return_hierarchy: if the hierarchy should be returned
        :return: the list of contours
        """
        if values is None:
            values = self.color
        mask = cv2.Canny(img, *values)

        return self.get_contours_mask(mask, save=save, return_hierarchy=return_hierarchy)

    def get_contours(self, img, color=None, save=True, return_hierarchy=False):
        """
        Action: Gets a list of all the contours within the range that
        NOTE: This method is only used for Color objects, for MultiColorRange please refer to the MultiColorRange
        class.
        :param color: a color object of the range to be detected
        :param img: image from which to get the contours
        :param save: if the contours returned from this function should be saved as self.contours
        :param return_hierarchy: if the hierarchy should be returned
        :return: list of all contours matching the range of hsv colours
        :rtype: list
        """
        if color is not None:
            c = color
        else:
            c = self.color
        # If there was a problem reading the image, exit
        if img is None:
            raise Exception("No picture given!")

        if type(img) is str:
            img = cv2.imread(img)

        if self.saturation_weight_vector is not None:
            image_mean = cv2.mean(img)[1]
            new_s = get_calibrated_value(image_mean, self.saturation_weight_vector)
            c.low[1] = new_s

        if self.value_weight_vector is not None:
            image_mean = cv2.mean(img)[2]
            new_v = get_calibrated_value(image_mean, self.value_weight_vector)
            c.low[2] = new_v

        img_mask = self.get_color_mask(img, color=c)
        return self.get_contours_mask(img_mask, save=save, return_hierarchy=return_hierarchy)

    def __filter_list(self):
        """
        Action: returns a __repr__ of the filter function list
        :return: string of the list of filter functions
        """
        return str([f.__name__ for f in self.__filters]).replace("'", '')

    def __morph_list(self):
        """
        Action: returns a __repr__ of the morphological function list
        :return: string of the list of morphological functions
        """
        return str([f.__name__ for f in self.morph]).replace("'", '')

    def __image_filter_list(self):
        """
        Action: returns a __repr__ of the image filters list
        :return: string of the list of image filter functions
        """
        return str([f.__name__ for f in self.image_filters]).replace("'", '')

    @staticmethod
    def get_match(func):
        """
        returns the matching function of the given function name (string)
        :param func: string of the name of a function
        :return: a function object by the name given
        """
        for m, n in Vision.list_of_functions:
            if isfunction(n):
                if str(func) == str(m):
                    return n

    def __repr__(self):
        """
        Action: Returns a string with the python code representing the Vision Object
        :return:
        """
        color = self.color.__repr__()
        filters = self.__filter_list()
        params = self.__params
        directions = self.directions.__name__
        first_part = numerical_replace('Vision(color=0, '
                                       'filters=1, '
                                       'directions_function=3, '
                                       'parameters=2, '
                                       'target_amount=4, '
                                       'width=5, '
                                       'height=6, '
                                       'connection_dst=7, '
                                       'port=8,'
                                       'camera_port=9, '
                                       'failed_value=#, '
                                       'swv=$,'
                                       'vwv=^,',
                                       color,
                                       filters,
                                       params,
                                       directions,
                                       self.target_amount,
                                       self.width,
                                       self.height,
                                       self.connection_address,
                                       self.network_port,
                                       self.camera_port).replace('#', str(self.failed_value))\
                                                        .replace('$', self.saturation_weight_vector.__repr__)\
                                                        .replace('^', self.value_weight_vector.__repr__)
        second_part = numerical_replace("image_filters=0,"
                                        "image_parameters=1,"
                                        "morph=2,"
                                        "morph_params=3)",
                                        self.__image_filter_list(),
                                        self.image_params,
                                        self.__morph_list(),
                                        self.morph_params)
        return first_part + second_part

    @staticmethod
    def __return_name(obj):
        if type(obj) == str:
            return obj
        elif obj is None:
            pass
        else:
            return obj

    def __str__(self):
        """
        Action: Recaps the data in the Vision object, useful for debugging
        :return: useful representation string of the Vision object
        """
        filters_for_print = (self.__return_name(i) for i in self.__filters)
        image_filters_for_print = (self.__return_name(i) for i in self.image_filters)
        morphs_for_print = (self.__return_name(i) for i in self.morph)

        filters = zip(filters_for_print, self.__params)
        image_filters = zip(image_filters_for_print, self.image_params)
        morphs = zip(morphs_for_print, self.morph_params)
        ct = type(self.color)
        if ct is MultiColor:
            color_type = 'Multiple Color Range (MultiColor)'
        elif ct is Color:
            color_type = 'Single Color Range Object (Color)'
        elif ct in (tuple, list, set):
            color_type = 'Canny edge detection (tuple, list, set)'
        else:
            color_type = 'Unknown'

        string = 'Vision Recap:\n' \
                 'Filters & custom Parameters: 0\n' \
                 'Color Detection Type: 1\n' \
                 'Color: 2\n' \
                 'Directions Function:8 \n' \
                 'Target Amount:9 \n' \
                 'Connection:\n' \
                 'Connection Type: \n' \
                 'Connection Address or IP: 3\n' \
                 'Connection Port: 4\n\n' \
                 'Camera:\n' \
                 'Camera port: 5\n' \
                 'Image Width & Height: (6, 7)\n'
        first_part = numerical_replace(string,
                                       filters,
                                       color_type,
                                       self.color.__repr__(),
                                       self.connection_address,
                                       self.network_port,
                                       self.camera_port,
                                       self.width,
                                       self.height,
                                       self.directions.__name__ if self.directions is not None else None,
                                       self.target_amount)
        string2 = 'Image filters: 0\n' \
                  'Morphological functions: 1\n' \
                  'Failed Value:2\n' \
                  'Log File:3\n' \
                  'Calibration JSON File:4'

        second_part = numerical_replace(string2,
                                        image_filters,
                                        morphs,
                                        self.failed_value,
                                        self.log_path,
                                        self.json_file)
        return first_part + second_part

    def get_directions(self, send=False, contours=None, amount=None, directions_function=None, sorter=None):
        """
        Action: Calculates the directions based on contours found
        :param send: A boolean denoting if to send the directions to the connection, destination
        :param contours: final contours after filtering
        :param amount: amount of final contours, this should be the amount of objects  you want to detect
        :param directions_function: optional parameter, a different directions function from the one defined
                                    in the Vision object.
        :param sorter: optional parameter,
        :return: a string of the directions (output of the directions function),
                 length depends on the directions function
        """
        if amount is None:
            amount = self.target_amount
        if contours is None:
            contours = self.contours
        if sorter is not None:
            target_contours = sorter(contours)[0:amount]
        else:
            target_contours = contours[0:amount]
        if directions_function is None:
            result = self.directions(target_contours, amount, (self.width, self.height))
        else:
            result = directions_function(target_contours, amount, (self.width, self.height))
        if send:
            self.send_to_destination(result)
        return result

    def camera_setup(self, port=0, img_width=None, img_height=None, camera=False):
        """
        Action: Opens up the camera reference and fixates a given width and height to all images taken
        :param img_width: the width of the images to be taken
        :param img_height: the height of the images to be taken
        :param name: name of thread of camera
        :param port: the camera port str for network connections int for local USB connections.
        :param camera: if the camera object should be ovl.Camera
        :return: the camera object, also sets self.camera to the object.
        """
        if img_height is None:
            img_height = self.height
        if img_width is None:
            img_width = self.width
        if type(port) not in [int, str]:
            raise TypeError("Invalid port type " + str(type(port)) + ' must be a string or an int.')
        self.camera_port = port
        if camera:
            robot_cam = Camera(src=port, width=img_width, height=img_height)
            robot_cam.start()
            # If there was a problem opening the camera, exit
            if robot_cam.isOpened() is False:
                raise Exception("An error has occurred! "
                                "cv2.VideoCapture returned None (Camera object unsuccessfully opened!")
        else:
            robot_cam = cv2.VideoCapture(port)
            # Set the width and height of the camera to the given parameters:
            if img_width != -1:
                robot_cam.set(3, img_width)
            if img_height != -1:
                robot_cam.set(4, img_height)
            if robot_cam.isOpened() is False:
                raise Exception("An error has occurred! "
                                "cv2.VideoCapture returned None (Camera object unsuccessfully opened!")
        self.camera = robot_cam
        return robot_cam

    def apply_sample(self, **kwargs):
        """
        Action: Finds contours and applies filters on a single image given by a image path, image object (numpy array)
        or camera port and takes an image, a custom color can be passed through the kw color.
        the kw display/show determines if the result should be displayed
        :param kwargs:
        :return: the image (numpy array) and the contours found (list of numpy arrays)
        """
        def cam_port_activate(port_val, color, wid, hgt):
            if self.camera is None:
                self.camera_setup(port_val, img_width=wid, img_height=hgt)
            return_val, image = self.camera.read()
            if return_val:
                self.contours = self.get_contours(image, color)
                self.apply_all_filters()
            else:
                print_pipe(self.log_path, "Something went wrong with getting the sample image!")
            return self.contours, image

        def image_activate(image_loc, color):
            if type(image_loc) is str:
                image = cv2.imread(image_loc, cv2.IMREAD_COLOR)
            else:
                image = image_loc
                self.contours = self.get_contours(image, color)
                self.apply_all_filters()
            return self.contours, image

        if "color" in kwargs:
            color_object = kwargs["color"]
        elif "hsv_low_limit" in kwargs and "hsv_high_limit" in kwargs:
            color_object = Color(kwargs["hsv_low_limit"], kwargs["hsv_high_limit"])
        else:
            color_object = self.color

        if "display" in kwargs:
            display = kwargs["display"]
        elif "show" in kwargs:
            display = kwargs["show"]
        else:
            display = None

        if "camera_port" in kwargs:
            port = kwargs["camera_port"]
            contours, img = cam_port_activate(port, color_object, self.width, self.height)
        elif "port" in kwargs:
            port = kwargs["port"]
            contours, img = cam_port_activate(port, color_object, self.width, self.height)
        elif "image" in kwargs:
            contours, img = image_activate(kwargs["image"], color_object)
        elif "img" in kwargs:
            contours, img = image_activate(kwargs["img"], color_object)
        elif "pic" in kwargs:
            contours, img = image_activate(kwargs["pic"], color_object)
        elif "picture" in kwargs:
            contours, img = image_activate(kwargs["picture"], color_object)
        else:
            raise ValueError("No image or camera port given!")

        if display:
            image_for_display = copy.copy(img)
            cv2.drawContours(image_for_display, contours, -1, BuiltInColors.red_high_hsv.high_bound)
            cv2.imshow(str(time.time()), image_for_display)
            cv2.waitKey()
        return contours, img

    class DoubleStack:
        def __init__(self):
            self.class1 = 0
            self.class2 = 0

        def inc1(self):
            self.class1 += 1

        def inc2(self):
            self.class2 += 1

        def bigger(self):
            if self.class1 > self.class2:
                return 1
            if self.class1 == self.class2:
                return 3
            return 2

        def sum(self):
            return self.class2 + self.class1

    def single_frame(self, port=0, apply_all=True):
        """
        Action: gets contours and applies all filters and returns the result
        :param port: port of the camera
        :param apply_all: if filters should be applied and have an effect on the contour list
        :return: contours, ratio list (from the filter functions)
        """
        if self.camera is None:
            self.camera_setup(port, self.width, self.height)
        return_val, new_img = self.camera.read()
        if return_val:
            new_img = self.apply_image_filters(new_img)
            if type(self.color) in (MultiColor, Color):
                self.contours = self.get_contours(new_img)
            else:
                self.contours = self.get_edge_contours(new_img)
            r = self.apply_all_filters(apply_all)
            return self.contours, r
        else:
            print_pipe(self.log_path, 'Failed to get image')

    def frame_loop(self, apply_all=True, one_loop=False, print_results=True,
                   send_direction=True, amend=True):
        """
        Action: frame loop: Takes an image, finds contours, applies all filters and sorters, finds direction and sends
                the direction to the RoboRIO in an infinite loop, amends for bad frames.
        :param apply_all: if filters should be applied
        :param one_loop: if loop should be executed once.
        :param print_results: if results should be printed through the pipe.
        :param send_direction: If directions should be sent to the RoboRIO.
        :param amend: If Bad frames should be amended by using previous values.
        :return: None
        """
        previous_result = None
        streak = Vision.DoubleStack()
        while True:
            old_time = time.time()
            for frame_num in xrange(30):
                return_val, new_img = self.camera.read()
                if return_val:
                    self.contours = self.get_contours(new_img)
                    self.apply_all_filters(apply_all=apply_all)
                    directions = self.directions(self.contours, self.target_amount, new_img)
                    if directions is False:
                        if previous_result is not None and amend:
                            directions = previous_result
                            streak.inc2()
                        elif previous_result is None:
                            self.send_to_destination(self.failed_value)
                            if print_results:
                                print_pipe(self.log_path, self.failed_value)
                            continue
                    if print_results:
                        print_pipe(self.log_path, directions)
                    if send_direction:
                        self.send_to_destination(directions)
                else:
                    if not one_loop:
                        print_pipe(self.log_path, "%d was an invalid picture!" % frame_num)
                    else:
                        print_pipe(self.log_path, "invalid picture!")
                    if previous_result is not None and amend:
                        streak.inc2()
                        directions = previous_result
                        if print_results:
                            print_pipe(self.log_path, directions)
                        if send_direction and self.connection_address is not None:
                            self.send_to_destination(directions)
                if streak.sum() > 3 and print_results:
                    res = streak.bigger()
                    if res is 1:
                        print_pipe(self.log_path,
                                   'Warning: The last {0} iterations have failed.'
                                   ' In {1} of them not enough contours were found to perform the directions function.'
                                   ''.format(streak.sum(), streak.class1))
                    if res is 2:
                        print_pipe(self.log_path,
                                   'Warning: The last {0} iterations have failed.'
                                   ' In {1} of them the camera failed to take an image.'
                                   ''.format(streak.sum(), streak.class2))
                    if res is 3:
                        print_pipe(self.log_path,
                                   'Warning: The last {0} iterations have failed.'
                                   ' In {1} of them the camera failed to take an image'
                                   ' and in {1} of them not enough contours were found.'
                                   ''.format(streak.sum(), streak.class2))
            if print_results:
                print_pipe(self.log_path, "Completed 30 frames in ", time.time() - old_time)
            if one_loop:
                break

    def apply_image_filters(self, img):
        """
        Action: applies all given image filters to the given image
        :param img: the image that the image filters should be applied on (numpy array)
        :return: the image with the filters applied
        """
        for idx, img_filter in enumerate(self.image_filters):
            img = img_filter(img, *self.image_params[idx])
        return img

    def start(self, **kwargs):
        """
        Action: Sets up the camera and activates the loop, find contours in an image, filter, get direction
        :param kwargs:
        :return: None
        """
        if "camera" in kwargs:
            port, width, height = kwargs["camera"]
        elif "cam" in kwargs:
            port, width, height = kwargs["cam"]
        else:
            port, width, height = 0, self.width, self.height
        if type(self.camera) is not cv2.VideoCapture:
            self.camera_setup(port, img_width=width, img_height=height)
        while True:
            self.frame_loop()

    def __call__(self, **kwargs):
        self.start(**kwargs)

    def translate_parameters(self):
        """
        Translates parameters to a json save-able format, capable of translating Color and MultiColor objects too.
        :return: the dictionary with the translated parameters for filter functions
        """
        param_copy = copy.copy(self.__params)
        for idx1, param_list in enumerate(param_copy):
            for idx2, val in enumerate(param_list):
                if type(val) is Color:
                    param_copy[idx1][idx2] = Color.json_serialize(val)
                if type(val) is MultiColor:
                    param_copy[idx1][idx2] = MultiColor.json_serialize(val)
        return param_copy

    def json_serialize(self, pass_camera=False, pass_network=False, pass_calibration=False):
        """
        Action: Translates the Vision object to a Json file for transportability
        :param pass_camera: a bool denoting if the camera parameters (camera port) should be serialized.
        :param pass_network: a bool denoting if the network parameters and connection information
                            (port, connection type and destination should be serialized as well.
        :param pass_calibration: a bool denoting if the calibration parameters (vectors and means) should be serialized.
        :return: the serialized string in json format.
        """
        if type(pass_network) is not bool:
            pass_network = False
        if type(pass_camera) is not bool:
            pass_camera = False
        if type(pass_calibration) is not bool:
            pass_calibration = False
        if type(self.color) is Color:
            color_object = Color.json_serialize(self.color)
        elif type(self.color) is MultiColor:
            color_object = MultiColor.json_serialize(self.color)
        else:
            raise TypeError('Color is not a color or MultiColor!')
        translation = {'color': color_object,
                       'filters': [f.__name__ for f in self.__filters],
                       'parameters': self.translate_parameters(),
                       'directions_function': self.directions.__name__,
                       'target_amount': self.target_amount,
                       'width': self.width,
                       'height': self.height,
                       'calibration_information': pass_calibration,
                       'camera_information': pass_camera,
                       'network_information': pass_network}
        if pass_camera:
            translation['camera_port'] = self.camera_port
        if pass_network:
            translation['port'] = self.network_port
            translation['destination'] = self.connection_address
            translation['connection_type'] = self.connection_type
        if pass_calibration:
            translation['v_vector'] = self.value_weight_vector
            translation['s_vector'] = self.saturation_weight_vector

        result = json.dumps(translation)
        return result

    @staticmethod
    def json_deserialize_secure(json_string):
        """
        NOTE: When opening json files acquired from the internet open in first with deserialize secure, if it doesn't
        return false, you can open it with json deserialize
        Action: Takes the content of a json file and turns opens it as a Vision Object
        :param json_string: A string of json data.
        :return: The dictionary containing the deserialized data (python objects)
        """
        deserialized = json.loads(json_string)
        d = deserialized
        if type(deserialized) is not dict:
            return False

        try:
            color = Color.deserialize(d['color'])
            if color is False:
                return False
            filters = [Vision.get_match(f) for f in d['filters'] if type(f) == str]
            params = d['parameters']
            for idx1, parameter_list in enumerate(params):
                for idx2, val in enumerate(parameter_list):
                    if type(val) is list:
                        params[idx1][idx2] = [Color.deserialize(c_object) for c_object in val if type(c_object) is dict]
                    elif type(val) is dict:
                        params[idx1][idx2] = Color.deserialize(val)
            directions = d['directions_function']
            if type(directions) is False:
                return False
            target_amount = d['target_amount']
            width = d['width']
            height = d['height']
            if type(width) is not int:
                return False
            if type(height) is not int:
                return False
            cam_port = None
            if d['camera_information'] is True:
                cam_port = d['camera_port']
                if not (type(cam_port) is int or cam_port is None):
                    return False
            roborio_name = 'NOCONNECTION'
            port = 9212
            nw = d['network_information']
            if nw is True:
                port = d['port']
                roborio_name = d['roborio_name']
                if type(port) is not int:
                    return False
                if type(roborio_name) != str:
                    return False
        except Exception as e:
            if e:
                pass
            return False
        if nw:
            return Vision(color=color,
                          filters=filters,
                          directions_function=directions,
                          port=port,
                          parameters=params,
                          camera_port=cam_port,
                          width=width,
                          height=height,
                          connection_dst=roborio_name,
                          target_amount=target_amount)
        return Vision(color=color,
                      filters=filters,
                      parameters=params,
                      target_amount=target_amount,
                      directions_function=directions,
                      camera_port=cam_port,
                      width=width,
                      height=height)

    @staticmethod
    def json_deserialize(json_string):
        """
        Action: Takes the content of a json file and turns opens it as a Vision Object
        :param json_string: A string of json data.
        :return: The dictionary containing the deserialized data (python objects)
        """
        deserialized = json.loads(json_string)
        d = deserialized
        color = Color.deserialize(d['color'])
        filters = [f for f in d['filters'] if type(f) == str]
        params = d['parameters']
        for idx1, parameter_list in enumerate(params):
            for idx2, val in enumerate(parameter_list):
                if type(val) is list:
                    params[idx1][idx2] = [Color.deserialize(color_obj) for color_obj in val if type(color_obj) is dict]
                elif type(val) is dict:
                    params[idx1][idx2] = Color.deserialize(val)
        directions = d['directions_function']
        target_amount = d['target_amount']
        width = d['width']
        height = d['height']
        cam_port = None
        if d['camera_information'] is True:
            cam_port = d['camera_port']
        roborio_name = 'NOCONNECTION'
        port = 9212
        nw = d['network_information']
        if nw is True:
            port = d['port']
            roborio_name = d['roborio_name']
        nw = d['network_information']
        if nw:
            return Vision(color=color,
                          filters=filters,
                          directions_function=directions,
                          parameters=params,
                          port=port,
                          camera_port=cam_port,
                          width=width,
                          height=height,
                          connection_dst=roborio_name,
                          target_amount=target_amount)
        return Vision(color=color,
                      filters=filters,
                      parameters=params,
                      target_amount=target_amount,
                      directions_function=directions,
                      camera_port=cam_port,
                      width=width,
                      height=height)

    def pack(self, file_path, pass_network, pass_camera, pass_calibration):
        """
        Action: serializes and writes the vision object (self) to
        :param file_path: The path to file where the object should be saved, should be a json file.
        :param pass_camera: a bool denoting if the camera parameters (camera port) should be serialized.
        :param pass_network: a bool denoting if the network parameters and connection information (port, connection type
                             and destination should be serialized as well.
        :param pass_calibration: a bool denoting if the calibration parameters (vectors and means) should be serialized.
        :return: True if save and serialize was succesful false otherwise
        """
        if file_path.endswith('.json'):
            with open(file_path, 'w') as f:
                f.write(self.json_serialize(pass_network=pass_network,
                                            pass_calibration=pass_calibration,
                                            pass_camera=pass_camera))
                f.close()
            return True
        else:
            return False

    NC_NUMERAL = 'numeral'
    NC_TIME = 'time'

    def display_contours(self, img, amount=0, color=None, save_path=None):
        """
        Action: Displays the image with the current list of contours
        :param img: image from which the contours were taken from, numpy array or image path
        :param color: the color of the of the contours outline
        :param amount: amount of contours to display
        :param save_path: if the image should be saved, pass the wanted result path
        :return: the image with the drawn contours.
        """
        if type(img) is str:
            img = cv2.imread(img)
        image_for_display = copy.copy(img)
        if self.contour_amount > 0:
            if amount > 0:
                output_contours = self.contours[0:amount]
            else:
                output_contours = self.contours

            if type(output_contours) is not list:
                output_contours = [output_contours]
            if color is None:
                color = (0, 0, 0)
            cv2.drawContours(image_for_display, output_contours, -1, color, 2)
        if type(save_path) is str:
            cv2.imwrite(save_path, image_for_display)
        display_image(image_for_display)
        return image_for_display

    def photo_array(self, amount, delay=3, name_convention=NC_TIME, return_images=False, path=False):
        """
        Action: Takes a series of images with a delay between shots and saves them.
        :param amount: the amount of images to take
        :param delay: the delay between images
        :param name_convention: the naming convention to save the pictures
        :param return_images: a boolean to denote if the image objects should be returned
        :param path: if images should be saved in a specific folder
        :return: if return images is true, returns the images
        """
        if self.camera is None:
            self.camera_setup()
        images = []
        image_num = 0
        while image_num < amount:
            ret, img = self.camera.read()
            if ret:
                images.append(img)
                if name_convention is Vision.NC_TIME:
                    save_name = 'image@ '.replace('@', str(image_num)) + \
                                str(time.strftime('%a%d%b%Y%H:%M:%S', time.gmtime()))
                elif name_convention is Vision.NC_NUMERAL:
                    save_name = 'image@'.replace('@', str(image_num))
                else:
                    save_name = str(name_convention) + str(image_num)
                if path:
                    save_name = str(path) + save_name
                cv2.imwrite(save_name, img)
                while True:
                    cv2.imshow('', img)
                    key = cv2.waitKey(delay)
                    if key == ord('y'):
                        amount += 1
                        break
                    elif key == ord('n'):
                        break
                    else:
                        pass

        if return_images:
            return images
        else:
            del images

    def rasses_vanegev(self, amount, delay=3, name_convention=NC_TIME, return_images=False, path=False):
        """
        Easter Egg Function :3 (identical to Vision.Vision.photo_array)
        Action: Takes a series of images with a delay between shots and saves them.
        :param self: a vision object
        :param amount: the amount of images to take
        :param delay: the delay between images
        :param name_convention: the naming convention to save the pictures
        :param return_images: a boolean to denote if the image objects should be returned
        :param path: if images should be saved in a specific folder
        :return: if return images is true, returns the images
        """
        return self.photo_array(amount, delay, name_convention, return_images, path)



# Copyright 2018 - 2019 Ori Ben-Moshe - All rights reserved.
from .General import root, radians2degrees, degrees2radians
import cv2
from math import pi, cos as cosinus, sin as sinus, log, tan as tangent, factorial, atan as arctan
from sys import version_info
import numpy as np


def atan(x):
    """
    Action: Arc tangent in degrees
    :param x: the tangent
    :return: the angle in degrees
    """
    return radians2degrees(arctan(x))


def y_intersection(line_slope, intercept, x_val):
    """
    Action: finds the y value of the line (y = mx + b) at point x
    :param line_slope: slope of the line (m)
    :param intercept: the intercept of the line (b)
    :param x_val: the value to be used (substituted with x)
    :return: the y value
    """
    return x_val, line_slope * x_val + intercept


def x_intersection(line_slope, intercept, y_val):
    """
    Action: calculates the x value of which the line according to the given y value
    :param line_slope:
    :param intercept:
    :param y_val:
    :return:
    """
    return (y_val - intercept) / float(line_slope), y_val if line_slope != 0 else (0, y_val)


def distance_from_frame(point, image_dimensions):
    """
    Action calculates the distance of a given point for the frame of the image based on the vector from the center
    and the point
    :param point: point (x,y tuple) or contour (ndarray)
    :param image_dimensions: the size of the image, (width, height)
    :return: the distance
    """
    p = get_contour_center(point) if type(point) is np.ndarray else point
    image_center = (image_dimensions[0] / 2 - .5, image_dimensions[1] / 2 - .5)
    s = slope(point, image_center)
    i = - s * p[0] + p[1]
    xs = (image_dimensions[0] - 1, 0)
    ys = (image_dimensions[1] - 1, 0)
    x_inter = x_intersection
    y_inter = y_intersection
    dbp = distance_between_points
    return min([dbp(x_inter(s, i, x), p) for x in xs] + [dbp(y_inter(s, i, y), p) for y in ys])


def horizon_angle(point, field_of_view, image_width):
    """
    Action: returns the angle compared to the center of the image for a given field of view image width and point
    :param point: the x val or center
    :param field_of_view: the horizontal field of view
    :param image_width: the image width
    :return: the angle compared to the center of the image , negative left positive right.
    """
    x_val = point[0] if type(point) in (tuple, list, set) else point
    return float(atan((x_val - (image_width - 1) / 2) / float(focal_length(image_width, field_of_view))))


def vertical_angle(point, field_of_view, image_height):
    """
    Action: returns the angle compared to the center of the image for a given field of view image height and point
    :param point: the x val or center
    :param field_of_view: the vertical field of view
    :param image_height: the image height
    :return: the angle compared to the center of the image , negative up positive down6.
    """
    y_val = point[1] if type(point) in (tuple, list, set) else point
    return float(atan((y_val - (image_height - 1) / 2) / float(focal_length(image_height, field_of_view))))


def focal_length(image_width, field_of_view):
    """
    Action: calculates the focal length given an image width and field of view
    :param image_width:
    :param field_of_view:
    :return: calculates the focal length, a float
    """
    return image_width / float((2 * tan(float(field_of_view / 2))))


def open_arc_length(contour):
    """
    Action: Returns the arc length of an "open" contour
    :param contour:
    :return:
    """
    return cv2.arcLength(contour, False)


def slope(p1, p2):
    """
    Action: returns the slope between 2 points
    :param p1: first point (x,y)
    :param p2: second point (x,y)
    :return: the slope
    """
    return  0. if p1[0] == p2[0] else ((float(p1[1]) - float(p2[1])) / (float(p1[0]) - float(p2[0])))


def angle_between_points(p1, p2):
    """
    Action: calculates the angle between the 2 points and the horizontal axis
    :param p1: the first point (2d)
    :param p2: the second point  (2d)
    :return: the angle
    """
    return radians2degrees(atan(slope(p1, p2)))


def line_angle(contour):
    """
    Action: returns the angle of the line created by the top-most and
    bottom-most points of a contour
    :param contour: the contour, a numpy array
    :return: the angle in degrees
    """
    if type(contour) != np.ndarray:
        raise TypeError("Contour must be a numpy array")
    topmost = tuple(contour[contour[:, :, 1].argmin()][0])
    bottommost = tuple(contour[contour[:, :, 1].argmax()][0])
    return angle_between_points(topmost, bottommost)


def get_approximation_open(contour, approximation_coefficient=0.02):
    """
    Action: returns the vertices of the approximation of the contour
    NOTE: for "open" contours!
    :param contour: open contour (ndarray)
    :param approximation_coefficient: approximation coefficient
    :return: list of vertices
    """
    arc = approximation_coefficient * cv2.arcLength(contour, False)
    return cv2.approxPolyDP(contour, arc, False)


def sin(x):
    """
    Action: The trigonometric function, sine.
    :param x: an angle in degrees
    :return: result, float
    """
    return sinus(degrees2radians(x))


def cos(x):
    """
    Action:  The trigonometric function, cosine.
    :param x: an angle in degrees
    :return: result, float
    """
    return cosinus(degrees2radians(x))


def tan(x):
    """
    Action: The trigonometric function, tangent.
    :param x: an angle in degrees
    :return: result, float
    """
    return tangent(degrees2radians(x))


math_funcs = {'sin': sin, 'cos': cos, 'tan': tan, 'fact': factorial, 'log': log}


def calculate_math_expression(equation, value, additional_functions=None, symbol='x'):
    """
    Action: calculates a math expression with a
    :param equation:
    :param value:
    :param additional_functions:
    :param symbol: The symbol in the equation
    :return:
    """
    expression = equation.replace(symbol, str(value))
    if additional_functions is None:
        additional_functions = math_funcs

    if version_info[0] == 3:
        exec('equation_res = ' + expression, additional_functions)
        return locals()['equation_res']
    else:
        locals().update(math_funcs)
        equation_res = 0
        exec('equation_res = ' + expression)
        return locals()['equation_res']


def get_circle_area(radius):
    """
    Action: calculates the area of a circle with the given radius.
    :param radius: The radius of the circle
    :return: Area of the circle
    """
    return pi * (radius ** 2)


def distance_between_points(p1, p2):
    """
    Action: Calculates the Distance between 2 points.
    :param p1: tuple of x and y value of point 1
    :param p2: tuple of x and y value of point 2
    :return: Float value of the distance between the 2 points
    :rtype: float
    """
    return root((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2, 2)


def get_fill_ratio_straight(contour, reverse_div=False):
    """
    Action: Returns the ratio between the contour and the straight bounding rectangle.
    :param contour: The contour to be compared
    :param reverse_div: If the rectangle's area should be divided by the contour.
    :return:  the ratio, the width of the bounding rectangle, the height of the bounding rectangle
    :rtype: float
    """
    _, _, w, h = cv2.boundingRect(contour)
    r_area = w * h
    c_area = cv2.contourArea(contour)
    if reverse_div:
        return float(r_area) / c_area, w, h
    return float(c_area) / r_area, w, h


def get_fill_ratio_triangle(contour, triangle=None, reverse_div=False):
    """
    Action: returns the ratio
    :param contour:
    :param triangle: convex hull for the minEnclosingTriangle function, can be ignored
    :param reverse_div: If the triangle's area should be divided by the contour.
    :return:
    """
    t_area, t = cv2.minEnclosingTriangle(contour, triangle)
    c_area = cv2.contourArea(contour)
    if reverse_div:
        return float(t_area) / c_area
    return c_area / float(t_area)


def get_fill_ratio_rotating(contour, reverse_div=False):
    """
    Action: Returns the ratio between the contour and the smallest bounding rectangle.
    :param contour: The contour to be compared
    :param reverse_div: If the Rectangle should be divided by the contour.
    :return: the ratio, the width of the bounding rectangle, the height of the bounding rectangle
    :rtype: float
    """
    rotated = cv2.minAreaRect(contour)
    r_area = rotated[1][0] * rotated[1][1]
    c_area = cv2.contourArea(contour)
    if reverse_div:
        return float(r_area) / c_area, rotated[1][0], rotated[1][1]
    return float(c_area) / r_area, rotated[1][0], rotated[1][1]


def get_fill_ratio_circle(contour, reverse_div=False):
    """
    Action: Returns the ratio between the contour and the smallest enclosing circle.
    :param contour: The contour to be compared
    :param reverse_div: If the circle should be divided by the contour.
    :return: the ratio, the radius of the enclosing circle
    :rtype: float
    """
    _, enclosing_radius = cv2.minEnclosingCircle(contour)
    circle_area = get_circle_area(enclosing_radius)
    contour_area = cv2.contourArea(contour)
    if reverse_div:
        return float(circle_area) / contour_area, enclosing_radius
    return float(contour_area) / circle_area, enclosing_radius


def get_contour_center(contour):
    """
    Action: Returns the coordinates center of a contour
    :param contour: a single contour
    :return: X coordinate of center, Y coordinate of center
    :rtype: int, int
    """
    moments = cv2.moments(contour)
    try:
        x, y = int(moments['m10'] / moments['m00']), int(moments['m01'] / moments['m00'])
    except ZeroDivisionError:
        raise ValueError('Contour given is too small!')
    return x, y


def n_polygon_angle(amount_of_sides):
    """
    Action: returns the inner angle of a polygon with the given amount of sides
    :param amount_of_sides: amount of sides the polygon has
    :return:
    """
    n = amount_of_sides
    return (n - 2) * (180 / n)


def get_approximation(contour, approximation_coef=0.02):
    """
    Action: returns the approximation
    :param contour:
    :param approximation_coef:
    :return:
    """
    peri = cv2.arcLength(contour, True)
    approx = cv2.approxPolyDP(contour, approximation_coef * peri, True)
    return approx


def n_polygon_area(length, amount_of_sides):
    """
    Action: area of a polygon of
    :param length:
    :param amount_of_sides:
    :return:
    """
    angle = 180 / amount_of_sides
    return 0.25 * amount_of_sides * (length ** 2) * (cos(angle)/sin(angle))


def circle_rating(contour, area_factor=0.9, radius_factor=0.8):
    """
    Action: returns a rating of how close is the circle to being a circle
    :param contour: the contour that its rating is calculated
    :param area_factor: the factor (p-value) that separates an area ratio that is a circle and one that isn't
    :param radius_factor: the factor (p-value) that separates an radius ratio that is a circle and one that isn't
    :return:
    """
    fill_ratio, radius = get_fill_ratio_circle(contour)
    _, _, w, h = cv2.boundingRect(contour)
    radius_ratio = root(((radius * 2) ** 2) / float(w) * h, 2)
    rating = (radius_ratio * radius_factor) * (fill_ratio * area_factor)
    return rating


def get_lengths_and_angles(contour, approximation_coefficient=0.02):
    """
    Action: returns a list of lengths and angles of a given contour.
    :param contour: contour (numpy array) to find its
    :param approximation_coefficient:
    :return: returns the list of sides
    """
    def cosine_sentence(p1, p2, p3):
        dbp = distance_between_points
        l1, l2, l3 = dbp(p1, p2), dbp(p1, p3), dbp(p2, p3)
        alpha = (l1**2 + l2**2 - l3**2) / (l1 * l2 * 2)
        return radians2degrees(alpha)

    approximation = get_approximation(contour, approximation_coefficient)
    vertices = []
    lengths = []
    angles = []
    for points in approximation:
        vertex = points[0][1], points[0][0]
        vertices.append(vertex)

    for idx, point in enumerate(vertices):
        if idx == len(approximation) - 1:
            lengths.append(distance_between_points(point, vertices[0]))
            angles.append(cosine_sentence(vertices[idx - 1], point, vertices[0]))
        else:
            lengths.append(distance_between_points(point, vertices[idx + 1]))
            angles.append(cosine_sentence(vertices[idx - 1], point, vertices[idx + 1]))
    return vertices, lengths, angles



# Copyright 2018-2019 Ori Ben-Moshe - All rights reserved.
import cv2
from . import Geometry
from .General import root
from sys import version_info
from .Geometry import distance_between_points as dbp, distance_from_frame as dff
from numpy import mean, std

if version_info[0] == 3:
    xrange = range


def image_center_filter(contour_list, image_dimensions=(320, 240), max_dist=0.7):
    """
    Action: filters out contours that their center is not close enough
            to the center of the image
    :param contour_list: a list of contours to be filtered
    :param image_dimensions: the size of the image(width, height)
    :param max_dist: the maximum percent of difference
    :return:
    """
    output = []
    ratio = []
    image_center = (image_dimensions[0] / 2 - .5, image_dimensions[1] / 2 - .5)
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        contour_center = Geometry.get_contour_center(current_contour)
        distance_ratio = dbp(contour_center, image_center) / float(dff(contour_center, image_dimensions))
        if distance_ratio >= max_dist:
            output.append(current_contour)
            ratio.append(distance_ratio)
    return output, ratio


def abs_distance_filter(contour_list, max_dist=50, min_dist=0, image_dimensions=(320, 240)):
    """
    Action: filters out contours that their center is not close enough
            to the center of the image
    :param contour_list: a list of contours to be filtered
    :param image_dimensions: the size of the image(width, height)
    :param max_dist: the maximum distance from the center in pixels
    :param min_dist: the minimum distance from the center in pixels
    :return:
    """
    output = []
    ratio = []
    image_center = (image_dimensions[0] / 2 - .5, image_dimensions[1] / 2 - .5)
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        contour_center = Geometry.get_contour_center(current_contour)
        distance_from_center = dbp(contour_center, image_center)
        if max_dist >= distance_from_center >= min_dist:
            output.append(current_contour)
            ratio.append(distance_from_center)
    return output, ratio


def length_filter(contour_list, min_length=50, max_length=76800):
    """
    Action: receives a list of contours and removes ones that are not long enough
            Note: for "open" contours!
    :param contour_list: list of contours (ndarray) to be filtered
    :param min_length: minimum length of a contour (in pixels)
    :param max_length: maximum length of a contour (in pixels)
    :return: list of filtered contours and list of the lengths
    """
    output = []
    ratio = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        perimeter = cv2.arcLength(current_contour, False)
        if min_length >= perimeter >= max_length:
            output.append(current_contour)
            ratio.append(perimeter)
    return output, ratio


def straight_square_filter(contour_list, min_area_ratio=0.8, min_len_ratio=0.95):
    """
    Action: receives a list of contours
    and returns only the ones that are approximately square
    Relation checked is [minimum ratio < (Circle radius / width * height * (square root of 2)) < maximum ratio]
    :param contour_list: List of Contours to filter
    :type contour_list: List
    :param min_len_ratio: maximum ratio between radius and sides of bounding rectangle
    :param min_area_ratio: the minimum ratio between the area of the contour and the area of the bounding shape
    :return: the contour list filtered.
     """
    output_list = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for currentContour in contour_list:
        fill_ratio, w, h = Geometry.get_fill_ratio_straight(currentContour)
        peri = cv2.arcLength(currentContour, True)
        approx = cv2.approxPolyDP(currentContour, 0.02 * peri, True)
        if fill_ratio > min_area_ratio and len(approx) == 4:
            _, enclosing_radius = cv2.minEnclosingCircle(currentContour)
            p1 = root(2, 2) * root(w * h, 2)
            p2 = 2 * enclosing_radius
            radius_ratio = p2 / p1
            if radius_ratio >= min_len_ratio:
                output_list.append(currentContour)
                ratio_list.append(fill_ratio * radius_ratio)
    return output_list, ratio_list


def rotated_square_filter(contour_list, min_area_ratio=0.8, min_ratio=0.95, max_ratio=1.05):
    """
    Action: receives a list of contours
    and returns only the ones that are approximately square
    Relation checked is [minimum ratio < (Circle radius / width * height * (square root of 2)) < maximum ratio]
    :param contour_list: List of Contours to filter
    :param max_ratio: maximum ratio between radius and sides of bounding rectangle
    :type max_ratio: float
    :param min_ratio: minimum ratio between radius and sides of bounding rectangle
    :type min_ratio: float
    :param min_area_ratio: minimum ratio between the area of the contours and the bounding shape
    :return: the contour list filtered.
     """
    ratio_list = []
    output_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for currentContour in contour_list:
        fill_ratio, w, h = Geometry.get_fill_ratio_rotating(currentContour)
        _, enclosing_radius = cv2.minEnclosingCircle(currentContour)
        peri = cv2.arcLength(currentContour, True)
        approx = cv2.approxPolyDP(currentContour, 0.02 * peri, True)
        if fill_ratio > min_area_ratio and len(approx) == 4:
            p1 = root(2, 2) * root(w * h, 2)
            p2 = 2 * enclosing_radius
            radius_ratio = p2 / p1
            if min_ratio < radius_ratio < max_ratio:
                output_list.append(currentContour)
                if radius_ratio > 1:
                    radius_ratio = 1 / radius_ratio
                ratio_list.append(fill_ratio * radius_ratio)
    return output_list, ratio_list


def vertical_rectangle_filter(contour_list, min_area_ratio=0.):
    """
    Action: receives a list of contours
    and returns only the ones that are approximately a vertical rectangle
    :param contour_list: List of Contours to filter
    :type contour_list: List
    :param min_area_ratio: minimum ratio between the area of the contours and the bounding shape
    :return: the contour list filtered.
     """
    ratio_list = []
    output_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for currentContour in contour_list:
        fill_ratio, w, h = Geometry.get_fill_ratio_straight(currentContour)
        peri = cv2.arcLength(currentContour, True)
        approx = cv2.approxPolyDP(currentContour, 0.02 * peri, True)
        if fill_ratio > min_area_ratio and w < h and len(approx) == 4:
            output_list.append(currentContour)
            ratio_list.append(fill_ratio)
    return output_list, ratio_list


def horizontal_rectangle_filter(contour_list, min_area_ratio=0.8):
    """
    Action: receives a list of contours and returns only those that are approximately a horizontal rectangle
    :param contour_list: List of Contours to filter
    :type contour_list: List
    :param min_area_ratio
    :return: the contour list filtered.
     """
    output_list = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for currentContour in contour_list:
        fill_ratio, w, h = Geometry.get_fill_ratio_straight(currentContour)
        peri = cv2.arcLength(currentContour, True)
        approx = cv2.approxPolyDP(currentContour, 0.02 * peri, True)
        if fill_ratio > min_area_ratio and w > h and len(approx) == 4:
            output_list.append(currentContour)
            ratio_list.append(fill_ratio)
    return output_list, ratio_list


def area_filter(contour_list, min_area=200, max_area=76800):
    """
    Action: filters contours that are not within the threshold of area

    :param max_area: maximum area of a contour (Inclusive) set to the size of the image for no upper limit.
    :param min_area: minimum area of a contour (Inclusive) set to 0 for no lower limit.
    :param contour_list: List of Contours to filter
    :type contour_list: List
    :return: the contour list filtered.
    """
    output_list = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        if min_area <= cv2.contourArea(current_contour) <= max_area:
            output_list.append(current_contour)
            ratio_list.append(current_contour)
    return output_list, ratio_list


def straight_rectangle_filter(contour_list, min_area_ratio=0.8):
    """
    Action: receives a list of contours and returns only those that are approximately a rectangle
    and that are not rotated

    :param contour_list: List of Contours to filter
    :type contour_list: List
    :param min_area_ratio: The minimum ratio between the rectangle and the contour
    :type min_area_ratio: float
    :return: the contour list filtered.
     """
    output_list = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for currentContour in contour_list:
        fill_ratio, _, _ = Geometry.get_fill_ratio_straight(currentContour)
        peri = cv2.arcLength(currentContour, True)
        approx = cv2.approxPolyDP(currentContour, 0.02 * peri, True)
        if fill_ratio > min_area_ratio and len(approx) == 4:
            output_list.append(currentContour)
            ratio_list.append(fill_ratio)
    return output_list, ratio_list


def rotated_rectangle_filter(contour_list, min_area_ratio):
    """
    Action: receives a list of contours and returns only those that are approximately a rectangle regardless
    of the angle of rotation.
    :param contour_list: List of Contours to filter
    :type contour_list: List
    :param min_area_ratio: the minimum ratio between the contour area and the bounding shape
    :return: the contour list filtered.
     """
    output_list = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for currentContour in contour_list:
        fill_ratio, w, h = Geometry.get_fill_ratio_rotating(currentContour)
        peri = cv2.arcLength(currentContour, True)
        approx = cv2.approxPolyDP(currentContour, 0.02 * peri, True)
        if fill_ratio > min_area_ratio and len(approx) == 4:
            ratio_list.append(fill_ratio)
            output_list.append(currentContour)
    return output_list, ratio_list


def triangle_filter(contour_list, min_area_ratio=0.8):
    """
    Action: receives a list of contours and returns only those that are approximately
    traingle
    :param contour_list:
    :param min_area_ratio:
    :return:
    """
    output_list = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for currentContour in contour_list:
        fill_ratio = Geometry.get_fill_ratio_triangle(currentContour)
        peri = cv2.arcLength(currentContour, True)
        approx = cv2.approxPolyDP(currentContour, 0.02 * peri, True)
        if fill_ratio > min_area_ratio and len(approx) == 3:
            ratio_list.append(fill_ratio)
            output_list.append(currentContour)
    return output_list, ratio_list


def circle_filter(contour_list, min_area_ratio=0.80, min_len_ratio=0.9):
    """
    Action: filters out contour which are not approximately circle.
    :param contour_list: list of arrays (numpy uint8 ndarrays (contours) to be filtered
    :param min_area_ratio: minimum ratio between the area of the enclosing
                      circle and the contour (contour /enclosing circle)
    :param min_len_ratio: minimum ratio between the radius of the enclosing circle and the enclosing rectangle.
    :return: the list of contours that fit the criteria
    """
    output = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for currentContour in contour_list:
        fill_ratio, radius = Geometry.get_fill_ratio_circle(currentContour)
        _, _, w, h = cv2.boundingRect(currentContour)
        radius_ratio = ((2 * radius) ** 2) / float(w) * h
        if min_len_ratio <= root(radius_ratio, 2):
            if min_area_ratio <= fill_ratio:
                output.append(currentContour)
                ratio_list.append((fill_ratio, radius_ratio))
    return output, ratio_list


def regular_polygon_filter(contour_list,
                           side_amount=6,
                           min_angle_ratio=0.9,
                           min_area_ratio=0.85,
                           min_len_ratio=0.9,
                           approx_coef=0.02):
    """
    Action: A filter that Detects regular polygon of n-sides sides
    :param contour_list: the list of contours to be filtered
    :param side_amount: the amount of sides the wanted polygon has.
    :param min_angle_ratio: the minimum ratio between the each angle and the average angle and the average angle and the
                            target angle of a shape of side_amount
    :param min_area_ratio: The minimum ratio between the contour's area and the target area
    :param min_len_ratio: The minimum ratio between the length of each side and the average length and
                          between the average length.
    :param approx_coef: the coefficient for the function cv2.approxPolyDP.
    :return:
    """
    if side_amount < 3:
        return []
    output = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for currentContour in contour_list:
        vertices, lengths, angles = Geometry.get_lengths_and_angles(currentContour, approx_coef)
        if len(vertices) == side_amount:
            target_angle = Geometry.n_polygon_angle(side_amount)
            average_length = round(sum(lengths) / float(len(lengths)), 3)
            average_angle = round(sum(angles) / float(len(lengths)), 3)
            invalid_value = False
            for length in lengths:
                if 1 - (root((length - average_length)**2, 2) / average_length) < min_len_ratio:
                    invalid_value = True
                    break
            if not invalid_value:
                for angle in angles:
                    if 1 - (root((angle - average_angle)**2, 2) / average_angle) < min_angle_ratio:
                        invalid_value = True
                        break
                if not invalid_value:
                    if 1 - (root((target_angle - average_angle)**2, 2) / target_angle) < min_angle_ratio:
                        invalid_value = True
                    if not invalid_value:
                        ratio = cv2.contourArea(currentContour) / Geometry.n_polygon_area(average_length, side_amount)
                        if min_area_ratio <= ratio <= 1 / min_area_ratio:
                            ratio_list.append(ratio)
                            output.append(currentContour)
    return output, ratio_list



# Copyright 2018-2019 Ori Ben-Moshe - All rights reserved.
from numpy import ndarray
from . import Geometry
from sys import version_info


if version_info[0] == 3:
    xrange = range


def validate(contour, target_amount):
    if contour is None:
        return False
    elif contour is []:
        return False
    elif type(contour) is ndarray and target_amount > 1:
        return False
    elif type(contour) is list and len(contour) < target_amount:
        return False
    else:
        return True


def alert_directions(contour, target_amount, img_size=(320, 240)):
    """
    Action: returns true if targets were found false otherwise
    :param contour:
    :param target_amount:
    :param img_size:
    :return:
    """
    return 1 if validate(contour, target_amount) else 0


def xy_center_directions(contour, target_amount, img_size=(320, 240), invert=False):
    """
    Action: returns the directions for the robot for the given contours for the x and y axis
    :param contour: the final contours - your targets
    :param target_amount: the amount of targets wanted
    :param img_size: the size of the image (width, height)
    :param invert: if the value should be inverted (max - value)
    :return: the directions
    """
    width, height = img_size
    x_ratio = 2000 / width
    y_ratio = 2000 / height
    if not validate(contour, target_amount):
        return False
    if type(contour) is ndarray:
        contour = [contour]
    x_sum = 0
    y_sum = 0
    contour = contour[:target_amount]
    for currentContour in contour:
        x, y = Geometry.get_contour_center(currentContour)
        x_sum += x
        y_sum += y
    x_res = x_sum / len(contour)
    y_res = y_sum / len(contour)
    x_res *= x_ratio
    y_res *= y_ratio
    if invert:
        return str(2000 - x_res) + str(2000 - y_res)
    return str(x_res) + str(y_res)


def y_center_directions(contour, target_amount, img_size=(320, 240), invert=False):
    """
    Action: returns the directions for the robot for the given contours for the y axis only
    :param contour: the final contours - your targets
    :param target_amount: the amount of targets wanted
    :param img_size: the size of the image (width, height)
    :param invert: if the value should be inverted (max - value)
    :return: the directions
    """
    _, height = img_size
    y_ratio = 2000 / height
    if not validate(contour, target_amount):
        return False
    if type(contour) is ndarray:
        contour = [contour]
    y_sum = 0
    contour = contour[:target_amount]
    for currentContour in contour:
        x, y = Geometry.get_contour_center(currentContour)
        y_sum += y
    y_res = y_sum / len(contour)
    if invert:
        return 2000 - (y_res * y_ratio)
    return y_res * y_ratio


def x_center_directions(contour, target_amount, img_size=(320, 240), invert=False):
    """
    Action: returns the directions for the robot for the given contours for the x axis only
    :param contour: the final contours - your targets
    :param target_amount: the amount of targets wanted
    :param img_size: the size of the image (width, height)
    :param invert: if the value should be inverted (max - value)
    :return: the directions
    """
    width, _ = img_size
    x_ratio = 2000 / width
    if not validate(contour, target_amount):
        return False
    if type(contour) is ndarray:
        contour = [contour]
    x_sum = 0
    contour = contour[:target_amount]
    for currentContour in contour:
        x, y = Geometry.get_contour_center(currentContour)
        x_sum += x
    x_res = x_sum / len(contour)
    if invert:
        return 2000 - (x_res * x_ratio)
    return x_res * x_ratio




# Copyright 2018-2019 Ori Ben-Moshe - All rights reserved.
import json
from numpy import ndarray, array
import copy
import cv2
from . import General
from sys import version_info

if version_info[0] is 3:
    xrange = range


class Color(object):

    @staticmethod
    def json_serialize(color_object):
        translate = {'high': color_object.high_bound,
                     'low': color_object.low_bound,
                     'color_space': color_object.color_space}
        return translate

    def pack(self):
        return json.dumps(Color.json_serialize(self))

    @property
    def low_h(self):
        return self.low[0]

    @property
    def low_s(self):
        return self.low[1]

    @property
    def low_v(self):
        return self.low[2]

    @property
    def high_h(self):
        return self.high[0]

    @property
    def high_s(self):
        return self.high[1]

    @property
    def high_v(self):
        return self.high[2]

    @staticmethod
    def unpack(json_string):
        d = json.loads(json_string)
        try:
            if type(d['sensitivity']) not in [int, float]:
                return False
            if not Color.assert_hsv(d['high']):
                return False
            if not Color.assert_hsv(d['low']):
                return False
        except TypeError:
            return False
        except ValueError:
            return False
        except KeyError:
            return False
        return Color(low=d['low'], high=d['high'])

    @staticmethod
    def deserialize(color_dict):
        d = color_dict
        try:
            if type(d['sensitivity']) not in [int, float]:
                return False
            if not Color.assert_hsv(d['high']):
                return False
            if not Color.assert_hsv(d['low']):
                return False

        except TypeError:
            return False
        except ValueError:
            return False
        except KeyError:
            return False
        return Color(low=d['low'], high=d['high'])

    def __repr__(self):
        return 'Color(' + str(self.low).replace("'", '') + ', ' + str(self.high).replace("'", '') + ')'

    def copy(self):
        return copy.copy(self)

    @property
    def low_bound(self):
        if type(self.__low_bound) is ndarray:
            return self.__low_bound.tolist()
        return self.__low_bound

    @property
    def high_bound(self):
        if type(self.__high_bound) is ndarray:
            return self.__high_bound.tolist()
        return self.__high_bound

    @staticmethod
    def validate(val, ceiling):
        """
        Action: Checks if val is positive and less than the ceiling value.
        :param val: The value to be checked (usually an int)
        :param ceiling: The highest value "val" can be. (usually an int)
        :return: True if val is less than ceiling and not negative
        """
        if ceiling <= 0:
            raise ValueError("Ceiling needs to be a positive value larger than 0.")
        val = int(val)
        return 0 <= val < ceiling

    @staticmethod
    def assert_hsv(hsv_point):
        """
        Action: Makes sure the hsv point(or vector) given in the parameter has valid values.
        :param hsv_point: a 3 length list with 3 ints describing a point the hsv color space
        that describe a color limit in a range for a findContour function.
        :return: True if valid False if not.
        :rtype: bool
        """
        return Color.validate(hsv_point[0], 180) and Color.validate(hsv_point[1], 256) and Color.validate(hsv_point[2],
                                                                                                          256)

    @staticmethod
    def apply_vector(point, vector, add=True):
        """
        Action: adds or subtracts a vector (value change) to a point (a hsv color limit).
        :param point: The hsv limit as a list of [h value, s value, v value].
        :param vector: The change to be made as a list of [h change, s change, v change].
        :param add: a boolean that when false subtracts the point and the vector instead of adding them.
        :return: the sum (or difference) of the point and the vector.
        :rtype: list
        """
        if add:
            result = list(map(int.__add__, point, vector))
            if Color.assert_hsv(result):
                return result
            raise ValueError("HSV limit out of range!")
        else:
            result = list(map(int.__sub__, point, vector))
            if Color.assert_hsv(result):
                return result
            raise ValueError("HSV limit out of range!")

    def apply_vector_low(self, vector, add=True):
        """
        Action: adds or subtracts a vector (value change) to the low limit of the Color (a hsv color limit).
        NOTE: AUTOMATICALLY UPDATES THE ATTRIBUTE IN THE COLOR OBJECT
        :param vector: The change to be made as a list of [h change, s change, v change].
        :param add: a boolean that when false subtracts the point and the vector instead of adding them.
        :return: the sum (or difference) of the point and the vector.
        :rtype: list
        """
        if add:
            result = list(map(int.__add__, self.__low_bound, vector))
            if Color.assert_hsv(result):
                return Color(result, self.__high_bound)
            raise ValueError("HSV limit out of range!")
        else:
            result = list(map(int.__sub__, self.__low_bound, vector))
            if Color.assert_hsv(result):
                return Color(result, self.__high_bound)
            raise ValueError("HSV limit out of range!")

    def apply_vector_high(self, vector, add=True):
        """
        Action: adds or subtracts a vector (value change) to the low limit of the Color (a hsv color limit).
        NOTE: AUTOMATICALLY UPDATES THE ATTRIBUTE IN THE COLOR OBJECT
        :param vector: The change to be made as a list of [h change, s change, v change].
        :param add: a boolean that when false subtracts the point and the vector instead of adding them.
        :return: the sum (or difference) of the point and the vector.
        :rtype: list
        """
        if add:
            result = list(map(int.__add__, self.__high_bound, vector))
            if Color.assert_hsv(result):
                return Color(self.__low_bound, result)
            raise ValueError("HSV limit out of range!")
        else:
            result = list(map(int.__sub__, self.__high_bound, vector))
            if Color.assert_hsv(result):
                return Color(self.__low_bound, result)
            raise ValueError("HSV limit out of range!")

    def set_low(self, new_range):
        point = self.copy()
        point = point.apply_vector_low(self.low_bound, False)
        return point.apply_vector_low(new_range)

    def set_high(self, new_range):
        point = self.copy()
        point = point.apply_vector_high(self.high_bound, False)
        return point.apply_vector_high(new_range)

    class ConstantError(TypeError):
        pass

    const_list = ["yellow_hsv",
                  'blue_hsv',
                  "orange_hsv",
                  'purple_hsv',
                  "white_hsv",
                  "black_hsv",
                  "blue_hsv",
                  "green_hsv",
                  "grey_hsv",
                  "red_low_hsv",
                  "teal_hsv",
                  "red_high_hsv"]

    def __init__(self, low, high):
        """
        Action: Constructor for the Color class
        :param high: high hsv limit of the color
        :param low: low hsv limit of the color
        """
       
        if type(low) is tuple:
            low = list(low)
        if type(low) is int:
            low = [low, 100, 100]
        if type(high) is tuple:
            high = list(high)
        if type(high) is int:
            high = [high, 255, 255]
        self.__low_bound = array(low)
        self.__high_bound = array(high)
        self.color_space = 'hsv'

    @property
    def low(self):
        """
        Action: Returns the low hsv limit of the color.
        :return: An uint8 numpy array with the low limit of the color.
        :rtype: uint8 numpy array
        """
        return self.__low_bound

    @property
    def high(self):
        """
        Action: Returns the high hsv limit of the color.
        :return: An uint8 numpy array with the high limit of the color.
        :rtype: uint8 numpy array
        """
        return self.__high_bound

    def light(self, amount=50):
        """
        Action: returns a copy of the color with a change in S low limit according to the given amount, default is 50
        NOTE: FOR HSV COLOR SPACE ONLY
        :return: the changed Color object
        """
        point = self.copy()
        point = point.apply_vector_low([0, -amount, 0])
        Color.assert_hsv(point.low)
        return point

    def dark(self, amount=50):
        """
        Action: returns a copy of the color with a change in S low limit according to the given amount, default is 50
        NOTE: FOR HSV COLOR SPACE ONLY
        :return: the changed Color object
        """
        point = self.copy()
        point = point.apply_vector_low([0, 0, -amount])
        Color.assert_hsv(point.low)
        return point

    def __str__(self):
        return str(self.low) + ' ' + str(self.high)


class MultiColor(object):

    def __init__(self, colors):
        """
        Action
        :param colors:
        """
        self.colors = colors
        for idx, r in enumerate(colors):
            if type(r) in (list, tuple):
                self.colors[idx] = Color(r[0], r[1])
            elif type(r) is Color:
                pass

    def pack(self):
        return json.dumps(MultiColor.json_serialize(self))

    @staticmethod
    def json_serialize(multi_color_object):
        return [Color.json_serialize(c) for c in multi_color_object.colors]

    def __str__(self):
        return ', '.join([str(color) for color in self.colors])

    def __repr__(self):
        return General.numerical_replace('MultiColor(0)', [repr(i) for i in self.colors]).replace("'", '')

    def adapt(self, img):
        for color in self.colors:
            color.adapt(img)

    @staticmethod
    def get_contours_for_range(img, color, vision):
        """
        Action: Gets a list of all the contours within the range that.
        :param color: a Color object describing (or numpy uint8 array) of color limit range in the hsv format.
        :param img: image from which to get the contours.
        :param vision: the vision object with a calibration to read from
        :return: list of all contours matching the range of hsv colours.
        :rtype: list, single numpy array (single contour) or None if no contour was found.
        """
        # If there was a problem reading the image, exit
        if img is None:
            raise Exception("No picture given!")
        hsv_low, hsv_high = color.low, color.high

        if vision.saturation_weight_vector is not None:
            image_mean = cv2.mean(img)[1]
            new_s = General.get_calibrated_value(image_mean, vision.saturation_weight_vector)
            hsv_low[1] = new_s

        if vision.value_weight_vector is not None:
            image_mean = cv2.mean(img)[2]
            new_v = General.get_calibrated_value(image_mean, vision.value_weight_vector)
            hsv_low[2] = new_v

        img_in_hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
        img_mask = cv2.inRange(img_in_hsv, hsv_low, hsv_high)

        if cv2.__version__.startswith("3."):
            return cv2.findContours(img_mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)[1]
        if cv2.__version__.startswith("4."):
            return cv2.findContours(img_mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)[1]
        elif cv2.__version__.startswith("2."):
            return cv2.findContours(img_mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)[0]
        else:
            raise General.VersionError("This code works with version 2 and 3 of openCV!")

    def get_contours(self, img, vision):
        """
        Gets contours for all the color ranges in the multi color range object from the given picture.
        :param img: image from which to detect the contours according to the color ranges
        :param vision: the vision object
        :return: a list of the contours found for all of the
        """
        contours = []
        for color in self.colors:
            result = MultiColor.get_contours_for_range(img, color=color, vision=vision)
            if type(result) is list:
                contours.extend(result)
            elif type(result) is ndarray:
                contours.append(result)
            else:
                pass
        return contours

    def light(self, amount=50):
        """
        Action: Uses the .light() function on all of the colors in the MultiColors.
        :param amount: The amount to be subtracted from each color's low saturation value
        :return: None
        """
        for idx, color_range in enumerate(self.colors):
            self.colors[idx] = color_range.light(amount)

    def dark(self, amount=50):
        """
        Action: Uses the .light() function on all of the colors in the MultiColors.
        :param amount: The amount to be subtracted from each color's low value value
        :return: None
        """
        for idx, color_range in enumerate(self.colors):
            self.colors[idx] = color_range.dark(amount)


class BuiltInColors(General.ToolClass):
    """
    A list of Built in colors for the use of colour based contour vision,
    can use as a base colour for a custom colour (using the apply hsv vector method of the
    color class or as a color on its own.
    All colours as static variables of the BuiltInColors class and can be accessed:
    BuiltInColors.<color name>_hsv
    The BuiltInColor class is a resource class with a bunch of built-in ready to use
    colors. These Colors can be used as a base for custom made colours using the Color
    class' method apply_hsv_vector (or one of the low or high specific functions)
    using a vector as a parameter to the method [h, s, v].
    List of colors:
      Red (MultiColorObject) : Red (low) + Red (high)
      Red (Low): [0, 100, 100], [8, 255, 255]
      Red (High): [172, 100, 100], [179, 255, 255]
      Note: in order to find red, use both ranges (low and high) and use the some of both results.
      Blue: [105, 100, 100], [135, 255, 255]
      Green: [45, 100, 100], [75, 255, 255]
      Yellow: [15, 100, 100], [45, 255, 255]
      Orange: [10, 100, 100], [20, 255, 255]
      Grey: [0, 0, 0], [179, 50, 195]
      Black: [0, 0, 0], [179, 255, 30]
      White: [0, 0, 200], [179, 20, 255]
      Teal: [110, 100, 100], [130, 255, 255]
      Purple: [135, 100, 100], [165, 255, 255]
    """
    yellow_hsv = Color([25, 100, 100], [55, 255, 255])
    blue_hsv = Color([100, 100, 100], [135, 255, 255])
    green_hsv = Color([35, 100, 100], [75, 255, 255])
    purple_hsv = Color([130, 100, 100], [170, 255, 255])
    orange_hsv = Color([10, 100, 100], [25, 255, 255])
    red_low_hsv = Color([0, 100, 100], [15, 255, 255])
    red_high_hsv = Color([170, 100, 100], [179, 255, 255])
    red_hsv = MultiColor([red_high_hsv, red_low_hsv])
    white_hsv = Color([0, 0, 200], [179, 20, 255])
    black_hsv = Color([110, 0, 0], [130, 255, 30])
    grey_hsv = Color([0, 0, 0], [179, 50, 195])
    teal_hsv = Color([80, 50, 60], [95, 255, 155])

    built_in_list = [yellow_hsv,
                     blue_hsv,
                     green_hsv,
                     grey_hsv,
                     purple_hsv,
                     orange_hsv,
                     red_hsv,
                     white_hsv,
                     black_hsv,
                     teal_hsv]




# Copyright 2018-2019 Ori Ben-Moshe - All rights reserved.
from math import pi
from time import sleep, gmtime, strftime
from sys import version_info
from json import dumps, loads


if version_info[0] == 3:
    xrange = range


def print_pipe(out, *args):
    """
    Action: outputs the given args to the output given, screen, file or not at all.
    :param out: the output location None to screen False doesnt print a path is saved to a file
    :param args: the data to be printed
    :return: None
    """
    if out is False:
        pass
    elif out is None:
        for i in args:
            print(str(i))
        print('')
    else:
        with open(out, 'a') as f:
            print(args)
            text = strftime("%a, %d %b %Y %H:%M:%S:: ", gmtime()) + str(args)
            f.write(text)
            f.close()


class ToolClass:
    help_text = "The help and documentation text for %s hasn't been defined"

    @property
    def help(self):
        return self.__class__.help_text.replace('%s', self.__class__.__name__)

    def __init__(self):
        print('')
        print_pipe(None, self.help)
        sleep(0.25)
        raise ToolError("Tool classes cannot be instantiated ")


class TimeQueue:
    def __init__(self, capacity):
        self.time = []
        self.value = []
        self.capacity = capacity

    def push(self, value, time):
        """
        Action:
        :param value:
        :param time:
        :return:
        """
        self.time.append(time)
        self.value.append(value)
        if len(time) >= self.capacity:
            self.deque()

    def deque(self):
        """
        Action: removes the oldest time and value
        :return:
        """
        del self.time[0]
        del self.value[0]

    def get_value_at_time(self, time):
        """
        Action:Gets the value closest to a given time stamp
        :param time:
        :return:
        """
        return self.time[min(enumerate(self.time), key=lambda i: abs(self.time[i] - time))[1]]


class InvalidCustomError(Exception):
    pass


class TimeTracker:

    def __init__(self, name):
        self.name = name
        self.__times = []

    @property
    def length(self):
        return len(self.__times)

    def add_time(self, value):
        if type(value) in [float, int]:
            self.__times.append(value)

    def get_average(self):
        """
        Action: calculates the average of the times.
        :return: the float representing the average
        """
        return sum(self.__times) / float(len(self.__times))

    def get_sum_of_squares(self):
        """
        Action: Calculates the sum of square deviations of the times.
        :return: float
        """
        return sum((x - self.get_average()) ** 2 for x in self.__times)

    def get_standard_deviation(self, dof=0):
        """
        Action: returns the
        :param dof: degrees of freedom, 0 population sd, 1 for sample mean
        :return:
        """
        length = self.length
        if length < 2:
            raise ValueError('Variance requires at least 2 times.')
        population_var = self.get_sum_of_squares() / (length - dof)
        return root(population_var)

    def sd(self, dof=0):
        return self.get_standard_deviation(dof)

    def avg(self):
        return self.get_average()

    def serialize(self):
        f = {'times': self.__times, 'name': self.name}
        return dumps(f)

    def pack(self, path):
        with open(path, 'w') as f:
            f.write(self.serialize())
            f.close()


class Performance:
    def __init__(self, names):
        self.trackers = {}
        for name in names:
            self.trackers[name] = TimeTracker(name)

    def add(self, name, value):
        self.trackers[name].add_time(value)

    def get_averages(self):
        return [i.avg() for i in self.trackers]


class VersionError(Exception):
    pass


class ToolError(Exception):
    pass


def root(base, degree=2):
    """
    Action: Raises base to the root of degree. [base^(1/degree)]
    :param base: The number to be raised
    :param degree: The root
    :return: the result
    :rtype: Float
    """
    return float(base) ** (1 / float(degree))


def get_calibrated_value(img_mean, vector):
    """
    Action: Solves the calibration equation that finds the optimal low bound value for
            the saturation and value.
    :param img_mean: the mean if the image of which
    :param vector: the dictionary containing the coefficients and group mean.
    :return: the optimal low bound
    """
    data_mean = vector['mean'][0]
    z_mean = data_mean[0] * vector['co1'] + data_mean[1] * vector['co2']
    return (z_mean - (img_mean * vector['co1'])) / vector['co2']


def radians2degrees(radians):
    """
        :param radians: the radians to be converted
        :return: returns the degrees equivalent of the given degrees
        """
    return radians * 180 / pi


def degrees2radians(degrees):
    """
    :param degrees: the degrees to be converted
    :return: returns the radian equivalent of the given degrees
    """
    return degrees * pi / 180


def numerical_replace(string, *args):
    for i in xrange(len(args)):
        string = string.replace(str(i), str(i) + '!%$#*^&#@^#%')

    for number, replacement in enumerate(args):
        string = string.replace(str(number) + '!%$#*^&#@^#%', str(replacement))
    return string


def enumertype(iterable):
    for obj in iterable:
        yield type(obj), obj



# Copyright 2018 - 2019 Ori Ben-Moshe - All rights reserved.
import cv2
from . import Geometry
from sys import version_info

if version_info[0] == 3:
    xrange = range


def image_center_sort(contour_list, image_dimensions=(320, 240)):
    """
    Action: sorts the contours from the closest to the center to the farthest.
    NOTE: it is important to area filter before image_center_sort
    :param contour_list: list of contours to be sorted
    :type contour_list: List or one contour (ndarray)
    :param image_dimensions: the dimensions of the image
    :return:
    """
    if type(contour_list) is not list:
        return [contour_list]
    image_center = ((image_dimensions[0] - 1) / 2, (image_dimensions[1] - 1) / 2)
    return sorted(contour_list,
                  key=lambda x: Geometry.distance_between_points(image_center, Geometry.get_contour_center(x)))


def dec_area_sort(contour_list):
    """
    Action: sorts the list of contours from the largest to the smallest based on area of the contour
    :param contour_list: List of Contours to be sorted
    :type contour_list: List or one contour (ndarray)
    :return: the contour list sorted.
    """
    if type(contour_list) is not list:
        return [contour_list]
    contour_list = sorted(contour_list, key=lambda x: cv2.contourArea(x))[::-1]
    return contour_list


def inc_area_sort(contour_list):
    """
    Action: sorts the list of contours from the smallest to the largest based on area of the contour
    :param contour_list: List of Contours to filter
    :type contour_list: List or one contour (ndarray)
    :return: the contour list sorted.
    """
    if type(contour_list) is not list:
        return [contour_list]
    contour_list = sorted(contour_list, key=lambda x: cv2.contourArea(x))
    return contour_list


def circle_sort(contour_list, area_limit=0.9, radius_limit=0.8):
    """
    Action: sorts the list of contours according to how much they are circle from most similar to least
            using the circle rating function.
    :param contour_list: list of Contours to filter
    :type contour_list: list or one contour(ndarray)
    :param area_limit: The area limit for the circle rating, look at Geometry.circle_rating
    :param radius_limit: the radius limit for the circle rating, look at Geometry.circle_rating
    :return: the list sorted by circle rating
    """
    if type(contour_list) is not list:
        return [contour_list]
    contour_list = sorted(contour_list, key=lambda x: Geometry.circle_rating(x, area_limit, radius_limit))
    return contour_list


def dec_length_sort(contour_list):
    """
    Action: sorts the list of contours from the largest to the smallest based on area of the contour
    :param contour_list: List of Contours to filter
    :type contour_list: List or one contour (ndarray)
    :return: the contour list sorted.
    """
    if type(contour_list) is not list:
        return [contour_list]
    contour_list = sorted(contour_list, key=Geometry.open_arc_length)[::-1]
    return contour_list


def inc_length_sort(contour_list):
    """
    Action: sorts the list of contours from the smallest to the largest based on area of the contour
    :param contour_list: List of Contours to filter
    :type contour_list: List or one contour (ndarray)
    :return: the contour list sorted.
    """
    if type(contour_list) is not list:
        return [contour_list]
    contour_list = sorted(contour_list, key=Geometry.open_arc_length)
    return contour_list



# Copyright 2018-2019 Ori Ben-Moshe - All rights reserved.
import cv2
import numpy as np


def validate_odd_size(size):
    """
    Action: validates that a kernel shape is of odd  ints and of size 2
    :param size: the shape (size) to be checked
    :return: doesnt raise an error if it's ok.
    """
    if type(size) not in (list, tuple):
        raise TypeError("Kernel size must be a tuple or list of 2 odd integers!")
    if len(size) != 2:
        raise ValueError("Kernel size must be a tuple or list of 2 odd integers!")
    if size[0] % 2 != 1 or size[1] % 2 != 1:
        raise ValueError("Kernel size must be 2 odd integers!")


def is_odd_size(size):
    """
    Action: validates that a kernel shape is of odd  ints and of size 2
    :param size: the shape (size) to be checked
    :return: doesnt raise an error if it's ok.
    """
    if type(size) not in (list, tuple):
        return False
    if len(size) != 2:
        return False
    if size[0] % 2 != 1 or size[1] % 2 != 1:
        return False
    return True


def validate_kernel(size):
    """
    Action: Validates the Size of a kernel
    :param size: the size (shape)
    :return:
    """
    if type(size) not in (list, tuple):
        raise TypeError("Kernel size must be a tuple or list!")
    if len(size) != 2:
        raise ValueError("Kernel size must be a tuple or list!")
    for i in size:
        if type(i) is not int:
            raise TypeError("Size must consist only of ints")


def cross_kernel(size):
    """
    Action: Returns a cross (ones in a cross) kernel for morphological functions
    Example of a (5,5) cross:
    | 0 0 1 0 0 |
    | 0 0 1 0 0 |
    | 1 1 1 1 1 |
    | 0 0 1 0 0 |
    | 0 0 1 0 0 |
    :param size:  a tuple of size 2 of 2 odd integers denoting the size of the kernel
    :return: the numpy.array of the cross shape
    """
    validate_odd_size(size)
    return cv2.getStructuringElement(cv2.MORPH_CROSS, ksize=size)


def rectangle_kernel(size):
    """
    Action: Returns a rectangle (all ones) kernel for morphological functions
    Example of a (5,5) rectangle:
    | 1 1 1 1 1 |
    | 1 1 1 1 1 |
    | 1 1 1 1 1 |
    | 1 1 1 1 1 |
    | 1 1 1 1 1 |
    :param size:  a tuple of size 2 of 2 odd integers denoting the size of the kernel
    :return: the numpy.array of the cross shape
    """
    return cv2.getStructuringElement(cv2.MORPH_RECT, ksize=size)


def ellipse_kernel(size):
    """
    Action: Returns an ellipse (ones in the shape of an ellipse) kernel for morphological functions
    Example of a (5,5) ellipse:
    | 0 0 1 0 0 |
    | 1 1 1 1 1 |
    | 1 1 1 1 1 |
    | 1 1 1 1 1 |
    | 0 0 1 0 0 |
    :param size: a tuple of size 2 of 2 odd integers denoting the size of the kernel
    :return:
    """
    validate_odd_size(size)
    return cv2.getStructuringElement(cv2.MORPH_ELLIPSE, ksize=size)


def horizontal_line_kernel(size):
    """
    Action: Returns an horizontal line (a horizontal line of ones) kernel for morphological functions
    Example of a (5,5) horizontal line:
    | 0 0 0 0 0 |
    | 0 0 0 0 0 |
    | 1 1 1 1 1 |
    | 0 0 0 0 0 |
    | 0 0 0 0 0 |
    :param size: a tuple of size 2 of 2 odd integers denoting the size of the kernel
    :return:
    """
    validate_odd_size(size)
    arr = np.zeros(size, dtype=np.uint8)
    arr[int((size[0] - 1) / 2), ] = 1
    return arr


def vertical_line_kernel(size):
    """
    Action: Returns a vertical line (a vertical line of ones) kernel for morphological functions
    Example of a (5,5) vertical line:
    | 0 0 1 0 0 |
    | 0 0 1 0 0 |
    | 0 0 1 0 0 |
    | 0 0 1 0 0 |
    | 0 0 1 0 0 |

    :param size:  a tuple of size 2 of 2 odd integers denoting the size of the kernel
    :return:
    """
    validate_odd_size(size)
    arr = np.zeros(size, dtype=np.uint8)
    arr[:, int((size[1] - 1) / 2)] = 1
    return arr


def convert_to_hsv(img):
    """
    Action: converts an image to hsv - Mainly for beginner use
    :param img: image to be converted
    :return: the converted image
    """
    return cv2.cvtColor(img, cv2.COLOR_BGR2HSV)


def sharpen_image(image, size=(3, 3)):
    """
    Action: Sharpens an image by preforming convolution it with a sharpening matrix
    :param image: the image (ndarray)
    :param size: the size of the sharpening matrix
    :return: the new sharpened image
    """

    validate_odd_size(size)
    kernel = np.ones(size)
    kernel *= -1
    kernel[int((size[0] - 1) / 2), int((size[1] - 1) / 2)] = kernel.size
    return cv2.filter2D(image, -1, kernel)


def adaptive_brightness(src, target_average=127, img_in_hsv=False):
    """
    Action: Changes the brightness of every pixel so that the average is the target average
    :param src: The image to be changed (Numpy array)
    :param target_average: the target average for the image
    :type target_average: int (int between 0 - 255)
    :param img_in_hsv: bool noting if the image is in hsv
    :return: a copy of the image changed
    """
    src = src.copy()
    if not img_in_hsv:
        src = cv2.cvtColor(src, cv2.COLOR_BGR2HSV)
    increase = cv2.mean(src)[2] - target_average
    vid = src[:, :, 2]
    vid = np.where(vid <= 255 - increase, vid + increase, 255)
    src[:, :, 2] = vid
    if not img_in_hsv:
        src = cv2.cvtColor(src, cv2.COLOR_HSV2BGR)
    return src


def change_brightness(src, change=20,  img_in_hsv=False):
    """
    Action: Changes the brightness of every pixel by the given amount
    :param src: The image to be changed (Numpy array)
    :param change: the change (integer)
    :type change: int
    :param img_in_hsv: bool noting if the image is in hsv
    :return: a copy of the image changed
    """
    src = src.copy()
    if not img_in_hsv:
        src = cv2.cvtColor(src, cv2.COLOR_BGR2HSV)
    vid = src[:, :, 2]
    vid = np.where(vid <= 255 - change, vid + change, 255)
    src[:, :, 2] = vid
    if not img_in_hsv:
        src = cv2.cvtColor(src, cv2.COLOR_HSV2BGR)
    return src


def rotate_by_angle(image, angle):
    """
    Action: rotates the given image by a given angle
    :param image:
    :param angle:
    :return:
    """
    (h, w) = image.shape[:2]
    center_xy = (w / 2, h / 2)
    rotation_matrix = cv2.getRotationMatrix2D(center_xy, -angle, 1.0)
    cos = np.abs(rotation_matrix[0, 0])
    sin = np.abs(rotation_matrix[0, 1])
    new_image_dimensions = (int((h * sin) + (w * cos)), int((h * cos) + (w * sin)))
    rotation_matrix[0, 2] += (new_image_dimensions[0] / 2) - center_xy[0]
    rotation_matrix[1, 2] += (new_image_dimensions[1] / 2) - center_xy[1]
    return cv2.warpAffine(image, rotation_matrix, new_image_dimensions)


def rotate90_left(image):
    """
    Action: return a copy of the image rotated 90 degrees to the left (counter-clockwise)
    :param image: numpy array, image to be rotated
    :return: a copy of the image rotated.
    """
    (h, w) = image.shape[:2]
    center = (w / 2, h / 2)
    rotation_matrix = cv2.getRotationMatrix2D(center, 90, 1.0)
    return cv2.warpAffine(image, rotation_matrix, (h, w))


def rotate90_right(image):
    """
    Action: return a copy of the image rotated 90 degrees to the right (clockwise)
    :param image: numpy array, image to be rotated
    :return:
    """
    (h, w) = image.shape[:2]
    center = (w / 2, h / 2)
    rotation_matrix = cv2.getRotationMatrix2D(center, 270, 1.0)
    return cv2.warpAffine(image, rotation_matrix, (h, w))


def rotate180(image):
    """
    Action: return a copy of the image rotated 180
    :param image:  numpy array, image to be rotated
    :return: a copy of the image rotated.
    """
    (h, w) = image.shape[:2]
    center = (w / 2, h / 2)
    rotation_matrix = cv2.getRotationMatrix2D(center, 180, 1.0)
    return cv2.warpAffine(image, rotation_matrix, (h, w))



from distutils.core import setup
setup(
    name='ovl',
    packages=['ovl'],
    version='0.2.4.1',
    license='apache-2.0',
    author='Ori Ben-Moshe',
    author_email='ovl.contact.help@gmail.com',
    description='A python Module for Object tracking vision for Robots',
    url='https://github.com/1937Elysium/Ovl-Python',
    install_requires=['numpy', 'sklearn', 'pynetworktables', 'scipy'],
    classifiers=['Development Status :: 4 - Beta',
                 'License :: OSI Approved :: Apache Software License',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 ]
     )